import 'dart:collection';
import 'dart:io' show Platform;
import 'package:flutter/material.dart';
import 'package:taboola_sdk/descriptors/device_descriptor.dart';
import 'package:taboola_sdk/classic/webview_manager.dart';
import 'package:taboola_sdk/classic/taboola_webview.dart';
import 'package:taboola_sdk/classic/taboola_classic_listener.dart';
import 'package:taboola_sdk/utils/logger.dart';
import 'package:flutter_custom_tabs/flutter_custom_tabs.dart';
import 'package:taboola_sdk/utils/webview_utils.dart';
import 'package:taboola_sdk/utils/text_utils.dart';

class TaboolaClassicUnit extends StatefulWidget {
  TaboolaClassicUnit._(
    this._publisherId,
    this._mode,
    this._placement,
    this._pageUrl,
    this._pageType,
    this._pageId,
    this._targetType,
    this._isFeed,
    this._taboolaStandardListener,
    this._unrecognizedPageExtraPropertiesManager, {
    Key? key,
    viewId,
    ScrollController? scrollController,
    bool? keepAlive,
  })  : this.scrollController = scrollController ??= ScrollController(),
        this._viewId = viewId ?? new DateTime.now().millisecondsSinceEpoch,
        this._keepAlive = keepAlive ?? false,
        super(
            key:
                key); //Note: '{} means scrollController is an optional argument, it allows publishers to set an external ScrollController from their scrolling parent. Otherwise, on default, we create an internal listener and ask the user to set it in their scrolling view.'

  factory TaboolaClassicUnit(
    String publisherId,
    String mode,
    String placement,
    String pageUrl,
    String pageType,
    String pageId,
    String targetType,
    bool isFeed,
    TaboolaClassicListener taboolaClassicListener,
    HashMap<String, dynamic> unrecognizedPageExtraPropertiesManager, {
    int? viewId,
    ScrollController? scrollController,
    bool? keepAlive,
  }) {
    return TaboolaClassicUnit._(
      publisherId,
      mode,
      placement,
      pageUrl,
      pageType,
      pageId,
      targetType,
      isFeed,
      taboolaClassicListener,
      unrecognizedPageExtraPropertiesManager,
      viewId: viewId,
      scrollController: scrollController,
      keepAlive: keepAlive,
    );
  }

  void setUnitExtraProps(HashMap<String, dynamic> extraProperties) {
    if (extraProperties.isEmpty) {
      return;
    }

    for (var property in extraProperties.entries) {
      var key = property.key;
      var value = property.value;
      if (TextUtils.isEmptyOrNull(key) ||
          (value is String && TextUtils.isEmptyOrNull(value))) {
        continue;
      }
      _unrecognizedUnitExtraProps[key] = value;
    }
  }

  final HashMap<String, dynamic> _unrecognizedUnitExtraProps =
      new HashMap<String, dynamic>();
  final String _publisherId;
  final String _mode;
  final String _placement;
  final String _pageUrl;
  final String _pageType;
  final String _pageId;
  final String _targetType;
  final bool _isFeed;
  final int _viewId;
  final bool _keepAlive;
  final TaboolaClassicListener _taboolaStandardListener;
  final HashMap<String, dynamic> _unrecognizedPageExtraPropertiesManager;
  final ScrollController scrollController;

  @override
  State<StatefulWidget> createState() => TaboolaClassicUnitState();
}

class TaboolaClassicUnitState extends State<TaboolaClassicUnit>
    with AutomaticKeepAliveClientMixin, WidgetsBindingObserver {
  late WebViewManager _webViewManager;
  late TaboolaWebView _taboolaWebView;

  double _webViewHeight = 100;
  bool _didCallScrollListener = false;
  GlobalKey _key = GlobalKey();

  @override
  void initState() {
    super.initState();
    widget.setUnitExtraProps(widget._unrecognizedPageExtraPropertiesManager);
    _webViewManager = WebViewManager(widget._unrecognizedUnitExtraProps);
    _taboolaWebView = TaboolaWebView(
      widget._publisherId,
      widget._placement,
      widget._mode,
      widget._pageType,
      widget._pageUrl,
      widget._pageId,
      widget._targetType,
      _webViewManager,
      widget._viewId,
    );

    _webViewManager.openItemBrowser = openItemBrowser;
    _webViewManager.didResize = taboolaStandardDidResize;
    _webViewManager.didShow = taboolaStandardDidShow;
    _webViewManager.didFailToLoad =
        widget._taboolaStandardListener.didFailToLoad;
    _webViewManager.didClickOnItem =
        widget._taboolaStandardListener.didClickOnItem;

    widget.scrollController.addListener(callScrollListener);
    WidgetsBinding.instance.addObserver(this);
  }

  @override
  void didUpdateWidget(TaboolaClassicUnit oldWidget) {
    super.didUpdateWidget(oldWidget);
    configureScrollController(oldWidget);
  }

  /// This method will mostly switch which ScrollController the TaboolaWebView will be working with (The internal one or the parent, external Widget one).
  void configureScrollController(TaboolaClassicUnit oldWidget) {
    if (oldWidget.scrollController != widget.scrollController) {
      oldWidget.scrollController.removeListener(
          callScrollListener); //TODO: check what happens if removeListener gets a null here
    }

    // Replace existing scroll listener
    widget.scrollController.removeListener(callScrollListener);
    widget.scrollController.addListener(callScrollListener);
  }

  @override
  void dispose() {
    WidgetsBinding.instance.removeObserver(this);
    widget.scrollController.removeListener(callScrollListener);
    super.dispose();
  }

  @override
  void didChangeMetrics() {
    // This is a fix for the bug that sometime the listener of the scroll doesnt work
    callScrollListenerIfListenerNotWorking();
  }

  void callScrollListener() {
    _didCallScrollListener = true;
    scrollListener();
  }

  void callScrollListenerIfListenerNotWorking() {
    if (!_didCallScrollListener) {
      scrollListener();
    }
  }

  @override
  Widget build(BuildContext context) {
    super.build(context);
    return ConstrainedBox(
      key: _key,
      constraints: BoxConstraints(maxHeight: _webViewHeight),
      child: _taboolaWebView,
    );
  }

  void taboolaStandardDidShow(String placement) {
    widget._taboolaStandardListener.didShow(placement);
  }

  void taboolaStandardDidResize(String placement, double height) {
    final screenSize = DeviceDescriptor.getScreenSize();
    if (screenSize == null || screenSize.isEmpty) {
      return;
    }
    double maxHeight = screenSize[1];
    if (maxHeight < height) {
      height = maxHeight;
    }

    if (_webViewHeight != height) {
      setState(() {
        _webViewHeight = height;
        widget._taboolaStandardListener.didResize(placement, _webViewHeight);
      });
    }
  }

  void scrollListener() {
    bool isScrollEnabled = DeviceDescriptor.getScreenSize() != null
        ? _webViewHeight == DeviceDescriptor.getScreenSize()![1]
        : false;

    _taboolaWebView.scrollListener(
        widget.scrollController, widget._isFeed, isScrollEnabled, context);
  }

  Future<void> openItemBrowser(String itemUrl) async {
    try {
      await launch(
        itemUrl,
        customTabsOption: WebviewUtils.ccTabConfig(),
        safariVCOption: WebviewUtils.safariViewControllerConfig(),
      );
    } catch (e) {
      // An exception is thrown if browser app is not installed on Android device.
      Logger.logException(e.toString());
    }
  }

  @override
  bool get wantKeepAlive => widget._keepAlive;
}
