import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:taboola_sdk/utils/webview_utils.dart';
import 'package:webview_flutter/webview_flutter.dart';
import 'package:taboola_sdk/classic/webview_manager.dart';
import 'package:taboola_sdk/classic/webview_gesture_recognizer.dart';
import 'package:taboola_sdk/utils/logger.dart';
import 'dart:io' show Platform;

// Import for Android features.
import 'package:webview_flutter_android/webview_flutter_android.dart';

// Import for iOS features.
import 'package:webview_flutter_wkwebview/webview_flutter_wkwebview.dart';

class TaboolaWebView extends StatefulWidget {
  final WebViewManager _webViewManager;
  late final WebViewController _webViewController;

  final String _publisherId,
      _placement,
      _mode,
      _pageType,
      _pageUrl,
      _pageId,
      _targetType;

  final int viewId;

  TaboolaWebView(
      this._publisherId,
      this._placement,
      this._mode,
      this._pageType,
      this._pageUrl,
      this._pageId,
      this._targetType,
      this._webViewManager,
      this.viewId);

  @override
  State<TaboolaWebView> createState() => _TaboolaWebViewState();

  void _sendNotifyExternalRects(BuildContext _buildContext) {
    final RenderObject? renderBox = _buildContext.findRenderObject();

    if (renderBox != null && renderBox is RenderBox) {
      try {
        //Originally meant to catch an assertion failure (but no functional issue): "Flutter: Failed assertion: 'attached': is not true."
        final position = renderBox.localToGlobal(Offset.zero);
        Size windowSize = MediaQuery.of(_buildContext).size;

        Size webViewSize =
            Size(windowSize.width, _buildContext.size?.height ?? 0.0);
        _webViewManager.webViewRect = position & webViewSize;
        _webViewManager.nativeWindowRect = Offset(0, 0) & windowSize;
        _webViewManager.notifyExternalRects();
      } catch (error) {
        Logger.log('_sendNotifyExternalRects() | Caught error: $error');
      }
    }
  }

  ///Returns the offset from screen top (and left?) when the WebView is close to screen top the offset.dy will return close to 0 result.
  Offset getPositionOnScreen(BuildContext _buildContext) {
    final RenderObject? renderBoxRed = _buildContext.findRenderObject();
    if (renderBoxRed != null && renderBoxRed is RenderBox) {
      return renderBoxRed.localToGlobal(Offset.zero);
    }

    return Offset(0, 0);
  }

  // there is a bug on iOS < 16.4, the feed is scrolled randomly to the middle, we fix this by scroll to top
  void fixIOSAutoScroll(bool isFeed) {
    if (Platform.isIOS) {
      _webViewController.getScrollPosition().then((webviewOffset) => {
            if (isFeed && webviewOffset.dy > 0)
              {
                _webViewController.scrollTo(0, 0),
              }
          });
    }
  }

  /// An update event called from outside this WebView. This is where screen measurements decide if to enable our scroll switch logic.
  ///
  /// _vertical - This is the vertical GestureRecognizer we implemcdent our ScrollSwitch logic in. It's a Taboola 'WebViewGestureRecognizer' object.
  void scrollListener(ScrollController controller, bool isFeed,
      bool canEnableScroll, BuildContext _buildContext) {
    if (!controller.hasClients) {
      Logger.log(
          "Scroll Controller is not attached to a scrolling parent yet.");
      //This might happen if controller logic is called while the relevant parent widget is still building. Source: https://stackoverflow.com/questions/52114535/scrollcontroller-not-attached-to-any-scroll-views
      return;
    }
    fixIOSAutoScroll(isFeed);

    if (isFeed && canEnableScroll) {
      if (controller.offset >= controller.position.maxScrollExtent &&
          !controller.position.outOfRange) {
        Logger.log("reach the bottom");
      }
    }
    _sendNotifyExternalRects(_buildContext);
  }
}

class _TaboolaWebViewState extends State<TaboolaWebView> {
  late Set<Factory<OneSequenceGestureRecognizer>> _savedFactoryRecognizers;
  late final String _initialUrl;
  late final WebViewWidget _webViewWidget;
  late WebViewGestureRecognizer _vertical;

  @override
  void initState() {
    super.initState();
    _vertical =
        WebViewGestureRecognizer(widget, context); //..onUpdate = (_) {};
    _savedFactoryRecognizers = [
      Factory<WebViewGestureRecognizer>(
        () => _vertical,
      )
    ].toSet();

    _initialUrl = WebviewUtils.buildWebviewUrl(
        widget.viewId,
        widget._publisherId,
        widget._mode,
        widget._pageType,
        widget._targetType,
        widget._pageId,
        widget._placement,
        widget._pageUrl);
    configWebViewControllerParams();

    _webViewWidget = WebViewWidget.fromPlatformCreationParams(
      params: createPlatformWebViewWidgetCreationParams(),
    );
    widget._webViewManager.controller = widget._webViewController;
    widget._webViewManager.taboolaWebViewDidScrollToTop =
        _taboolaWebViewDidScrollToTop;
    initWebViewWithURL();
  }

  @override
  void dispose() {
    removeWebview();
    super.dispose();
  }

  void initWebViewWithURL() {
    widget._webViewController
      ..setJavaScriptMode(JavaScriptMode.unrestricted)
      ..loadRequest(Uri.parse(_initialUrl))
          .then((value) => {widget._sendNotifyExternalRects(context)});
  }

  /*
  Previously, the webview was not cleared, resulting in an issue where an alert popup would appear on Android devices after switching to tabs.
  After clearing the webview in the dispose() function, this problem was resolved.
  For more details, refer to the following link: https://github.com/flutter/flutter/issues/119616
 */
  void removeWebview() {
    widget._webViewController
        .removeJavaScriptChannel(widget._webViewManager.flutterBridge);
    widget._webViewController.loadRequest(Uri.parse('about:blank'));
  }

  void configWebViewControllerParams() {
    late final PlatformWebViewControllerCreationParams
        webKitWebViewControllerCreationParams;
    if (WebViewPlatform.instance is WebKitWebViewPlatform) {
      webKitWebViewControllerCreationParams =
          WebKitWebViewControllerCreationParams(
        allowsInlineMediaPlayback: true,
      );
    } else {
      webKitWebViewControllerCreationParams =
          const PlatformWebViewControllerCreationParams();
    }

    widget._webViewController = WebViewController.fromPlatformCreationParams(
        webKitWebViewControllerCreationParams);
    if (widget._webViewController.platform is AndroidWebViewController) {
      AndroidWebViewController.enableDebugging(true);
      (widget._webViewController.platform as AndroidWebViewController)
          .setMediaPlaybackRequiresUserGesture(false);
    }
  }

  PlatformWebViewWidgetCreationParams
      createPlatformWebViewWidgetCreationParams() {
    PlatformWebViewWidgetCreationParams webViewWidgetParam =
        PlatformWebViewWidgetCreationParams(
      controller: widget._webViewController.platform,
      gestureRecognizers: _savedFactoryRecognizers,
    );

    if (WebViewPlatform.instance is WebKitWebViewPlatform) {
      webViewWidgetParam = WebKitWebViewWidgetCreationParams
          .fromPlatformWebViewWidgetCreationParams(
        webViewWidgetParam,
      );
    } else if (WebViewPlatform.instance is AndroidWebViewPlatform) {
      webViewWidgetParam = AndroidWebViewWidgetCreationParams
          .fromPlatformWebViewWidgetCreationParams(webViewWidgetParam,
              // Taboola Feed implementation is not handled well by WebView texture rendering implementation. Despite limiting the WebView size, the texture rendered keeps growing, causing a crash.
              // Therefore Taboola requires using 'Hybrid Composition' rendering mode. According to WebView.platform documentaiton "Setting `platform` doesn't affect [WebView]s that were already created."
              // Notice: It is known that 'Hybrid Composition' comes at a cost of some performance on Android versions below 10.
              displayWithHybridComposition: true);
    }
    return webViewWidgetParam;
  }

  ///This is called from MobileLoader (through WebViewManger)
  void _taboolaWebViewDidScrollToTop() {
    if (_vertical != null) {
      _vertical.onWebViewScrolledToContentTop();
    }
  }

  @override
  Widget build(BuildContext context) {
    Logger.log(_initialUrl);

    return _webViewWidget;
  }
}
