import 'dart:io' show Platform;
import 'package:flutter/widgets.dart';
import 'package:flutter/gestures.dart';
import 'dart:core';
import 'package:taboola_sdk/classic/taboola_webview.dart';
import 'package:taboola_sdk/utils/logger.dart';

class WebViewGestureRecognizer extends OneSequenceGestureRecognizer {
  late TaboolaWebView _taboolaWebView;
  late BuildContext _buildContext;

  WebViewGestureRecognizer(
      TaboolaWebView taboolaWebView, BuildContext buildContext) {
    _taboolaWebView = taboolaWebView;
    _buildContext = buildContext;
  }

  //Follow WebView global y scrolling value
  double yGlobal = 0;
  Offset dragDistance = Offset.zero;

  /// This is the only event that we managed to switch scrolling to web content
  @override
  void addPointer(PointerEvent event) {
    startTrackingPointer(
        event.pointer); //Allow handleEvent() to handle the logic.
  }

  @override
  String get debugDescription => 'customPan';

  @override
  void didStopTrackingLastPointer(int pointer) {}

  bool canWVContentScrollUp = false;
  bool didReceiveFromMLScrollToTop = false;
  double mOldPointY = 0, mOldPointX = 0, mDeltaX = 0, mDeltaY = 0;

  @override
  void handleEvent(PointerEvent event) async {
    dragDistance = dragDistance + event.delta;
    if (isWebViewAtScreenTop()) {
      bool horizontalScrolling = false;
      switch (event.runtimeType) {
        case PointerDownEvent:
          mOldPointY = event.position.dy;
          mOldPointX = event.position.dx;
          break;
        case PointerMoveEvent:
          mDeltaX = mOldPointX - event.position.dx;
          mDeltaY = mOldPointY - event.position.dy;
          horizontalScrolling =
              isHorizontalScrolling(mDeltaX, mDeltaY);
          break;
      }

      if (event.runtimeType != PointerMoveEvent) {
        //Don't switch scroll control when not scrolling (pointer moving). This was implemented to prevent a need for an extra down scroll gesture when moving from native into web-content scroll.
        return;
      }

      yGlobal = dragDistance.dy;
      int scrollDirection = getScrollingDirection(mDeltaY);
      if (scrollDirection == _SCROLL_DIRECTION_DOWN) {
        // This works since we run this method inside a condition that the WebView is already at screen top
        canWVContentScrollUp = true;
        //reset our flag for the next time we'll get webViewDidScrollToTop() from ML
        if (didReceiveFromMLScrollToTop) {
          didReceiveFromMLScrollToTop = false;
        }
      }

      if (horizontalScrolling) {
        resolve(GestureDisposition.rejected); // Native internal will win
      } else if (scrollDirection == _SCROLL_DIRECTION_UP &&
          !canWVContentScrollUp) {
        //DeltaY<0 is user trying to scroll up + WebView content cannot scroll up: return touch to Native.
        resolve(GestureDisposition.rejected); // Native internal will win
      } else {
        resolve(GestureDisposition.accepted); // WebView internal will win
//        canWVContentScrollUp = true;
      }
    }
  }

  static const int _SCROLL_DIRECTION_UP = 1;
  static const int _SCROLL_DIRECTION_NONE = 1;
  static const int _SCROLL_DIRECTION_DOWN = -1;

  int getScrollingDirection(double touchMovementDeltaY) {
    if (touchMovementDeltaY < 0) {
      return _SCROLL_DIRECTION_UP;
    } else if (touchMovementDeltaY > 0) {
      return _SCROLL_DIRECTION_DOWN;
    } else {
      return _SCROLL_DIRECTION_NONE;
    }
  }

  ///This event is eventually called from MobileLoader, when the WebView's content reached top
  void onWebViewScrolledToContentTop() {
    Logger.log('onWebViewScrolledToContentTop');
    // since on bounce affect we received form ML multiple time webViewDidScrollToTop() event (negative number) we need to check if we already received it once
    // You can refer to the code on ML project that notify us on scrollToTop here: updateScrollPosition() in TaboolaMobile.ts
    if (!didReceiveFromMLScrollToTop) {
      didReceiveFromMLScrollToTop = true;
      canWVContentScrollUp = false;
      // Set y position to 0 because the webView scrolled to top
      yGlobal = 0;
      dragDistance = Offset.zero;
    }
  }

  bool isWebViewAtScreenTop() {
    //TODO: try to compare to parent scrollView y offset (this will help with stuff like ActionBar, etc..)
    return _taboolaWebView.getPositionOnScreen(_buildContext).dy <= 1;
  }

  /// Previously, we were determining if `mDeltaX > mDeltaY` to identify horizontal scroll,
  /// but this approach was flawed. It misclassified diagonal scrolls, such as those at
  /// 45 degrees, as horizontal scrolls.
  ///
  /// To address this, we've introduced a tweak to avoid unintentional horizontal scroll detection.
  /// By increasing `mDeltaY * 2`, we ensure accurate detection even for scenarios involving
  /// carousels and PageViews.
  ///
  /// This adjustment helps us correctly differentiate between horizontal and diagonal scrolls.
  ///
  bool isHorizontalScrolling(double mDeltaX, double mDeltaY) {
    double mDeltaXAsb = mDeltaX.abs();
    double mDeltaYAsb = mDeltaY.abs();
    final int increaseDeltaYBy =
        2; // The factor by which we adjust mDeltaY for improved scroll detection accuracy.

    // kTouchSlop: "The distance a touch has to travel for the framework to be confident that the gesture is a scroll gesture". Source: constants.dart.
    bool isScrollGesture = mDeltaXAsb > kTouchSlop;
    bool isHorizontalScroll = mDeltaXAsb > (mDeltaYAsb * increaseDeltaYBy); // Determining if the scroll is primarily horizontal while accounting for the adjusted mDeltaY.


    if (isScrollGesture && isHorizontalScroll) {
      return true;
    }
    return false;
  }
}
